"""
TODO: remove this file and use the monkey module directly
"""
from geowatch.monkey import monkey_lightning
disable_lightning_hardware_warnings = monkey_lightning.disable_lightning_hardware_warnings
