"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/exceptions.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import TimeSampleError  # NOQA
