# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/evaluator.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/evaluator.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=None,
    serialized_pb=b'\n\x18\x64\x65\x65plab2/evaluator.proto\x12\x08\x64\x65\x65plab2\"\xe5\x05\n\x10\x45valuatorOptions\x12\x16\n\neval_steps\x18\x01 \x01(\x05:\x02-1\x12\x1b\n\reval_interval\x18\x02 \x01(\x05:\x04\x35\x30\x30\x30\x12#\n\x17\x63ontinuous_eval_timeout\x18\x03 \x01(\x05:\x02-1\x12\x1d\n\x0fuse_tf_function\x18\x04 \x01(\x08:\x04true\x12\x1b\n\x10stuff_area_limit\x18\x06 \x01(\x05:\x01\x30\x12\x1b\n\x10thing_area_limit\x18\x13 \x01(\x05:\x01\x30\x12\x33\n&transformer_class_confidence_threshold\x18\x14 \x01(\x02:\x03\x30.7\x12\'\n\x1apixel_confidence_threshold\x18\x15 \x01(\x02:\x03\x30.4\x12#\n\x16\x63\x65nter_score_threshold\x18\x07 \x01(\x02:\x03\x30.1\x12\x15\n\nnms_kernel\x18\x08 \x01(\x05:\x01\x33\x12\x1b\n\x0ekeep_k_centers\x18\t \x01(\x05:\x03\x34\x30\x30\x12\x1f\n\x10save_predictions\x18\n \x01(\x08:\x05\x66\x61lse\x12\x19\n\x11override_save_dir\x18\x0b \x01(\t\x12\x1b\n\x0fnum_vis_samples\x18\x0c \x01(\x05:\x02\x31\x30\x12#\n\x14save_raw_predictions\x18\r \x01(\x08:\x05\x66\x61lse\x12,\n\x13raw_panoptic_format\x18\x11 \x01(\t:\x0ftwo_channel_png\x12%\n\x17\x63onvert_raw_to_eval_ids\x18\x0e \x01(\x08:\x04true\x12!\n\x12\x61\x64\x64_flipped_images\x18\x05 \x01(\x08:\x05\x66\x61lse\x12\x17\n\x0b\x65val_scales\x18\x0f \x03(\x02\x42\x02\x10\x01\x12\x34\n&merge_semantic_and_instance_with_tf_op\x18\x10 \x01(\x08:\x04true\x12\"\n\x13\x64\x65tailed_ap_metrics\x18\x12 \x01(\x08:\x05\x66\x61lse'
)

_EVALUATOROPTIONS = _descriptor.Descriptor(
    name='EvaluatorOptions',
    full_name='deeplab2.EvaluatorOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='eval_steps', full_name='deeplab2.EvaluatorOptions.eval_steps', index=0,
            number=1, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=-1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='eval_interval', full_name='deeplab2.EvaluatorOptions.eval_interval', index=1,
            number=2, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=5000,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='continuous_eval_timeout', full_name='deeplab2.EvaluatorOptions.continuous_eval_timeout', index=2,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=-1,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_tf_function', full_name='deeplab2.EvaluatorOptions.use_tf_function', index=3,
            number=4, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='stuff_area_limit', full_name='deeplab2.EvaluatorOptions.stuff_area_limit', index=4,
            number=6, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='thing_area_limit', full_name='deeplab2.EvaluatorOptions.thing_area_limit', index=5,
            number=19, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='transformer_class_confidence_threshold',
            full_name='deeplab2.EvaluatorOptions.transformer_class_confidence_threshold', index=6,
            number=20, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.7),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='pixel_confidence_threshold', full_name='deeplab2.EvaluatorOptions.pixel_confidence_threshold',
            index=7,
            number=21, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.4),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='center_score_threshold', full_name='deeplab2.EvaluatorOptions.center_score_threshold', index=8,
            number=7, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='nms_kernel', full_name='deeplab2.EvaluatorOptions.nms_kernel', index=9,
            number=8, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=3,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='keep_k_centers', full_name='deeplab2.EvaluatorOptions.keep_k_centers', index=10,
            number=9, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=400,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_predictions', full_name='deeplab2.EvaluatorOptions.save_predictions', index=11,
            number=10, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='override_save_dir', full_name='deeplab2.EvaluatorOptions.override_save_dir', index=12,
            number=11, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='num_vis_samples', full_name='deeplab2.EvaluatorOptions.num_vis_samples', index=13,
            number=12, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=10,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='save_raw_predictions', full_name='deeplab2.EvaluatorOptions.save_raw_predictions', index=14,
            number=13, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='raw_panoptic_format', full_name='deeplab2.EvaluatorOptions.raw_panoptic_format', index=15,
            number=17, type=9, cpp_type=9, label=1,
            has_default_value=True, default_value=b"two_channel_png".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='convert_raw_to_eval_ids', full_name='deeplab2.EvaluatorOptions.convert_raw_to_eval_ids', index=16,
            number=14, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='add_flipped_images', full_name='deeplab2.EvaluatorOptions.add_flipped_images', index=17,
            number=5, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='eval_scales', full_name='deeplab2.EvaluatorOptions.eval_scales', index=18,
            number=15, type=2, cpp_type=6, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=b'\020\001', file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='merge_semantic_and_instance_with_tf_op',
            full_name='deeplab2.EvaluatorOptions.merge_semantic_and_instance_with_tf_op', index=19,
            number=16, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='detailed_ap_metrics', full_name='deeplab2.EvaluatorOptions.detailed_ap_metrics', index=20,
            number=18, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=39,
    serialized_end=780,
)

DESCRIPTOR.message_types_by_name['EvaluatorOptions'] = _EVALUATOROPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EvaluatorOptions = _reflection.GeneratedProtocolMessageType('EvaluatorOptions', (_message.Message,), {
    'DESCRIPTOR': _EVALUATOROPTIONS,
    '__module__': 'deeplab2.evaluator_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.EvaluatorOptions)
})
_sym_db.RegisterMessage(EvaluatorOptions)

_EVALUATOROPTIONS.fields_by_name['eval_scales']._options = None
# @@protoc_insertion_point(module_scope)
