"""
Note:
    This module has been moved to :mod:`kwgis.gis.geotiff`.
"""
from kwgis.gis.geotiff import geotiff_metadata, geotiff_crs_info, make_crs_info_object, geotiff_filepath_info, parse_landsat_product_id, walk_geotiff_products  # NOQA
