"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/plots.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import show_affinity_sample_process  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_dense_sample_indices  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_temporal_sample_indices  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import plot_temporal_sample  # NOQA
