"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/affinity.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import affinity_sample  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import make_soft_mask  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import _handle_degenerate_weights  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import hard_time_sample_pattern  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import soft_frame_affinity  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import hard_frame_affinity  # NOQA
