"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/time_kernel_grammar.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import MultiTimeKernelTransformer  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import _global_multi_time_kernel_parser  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import parse_multi_time_kernel  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import MULTI_TIME_KERNEL_GRAMMAR  # NOQA
