"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/utils.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import guess_missing_unixtimes  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import coerce_time_kernel  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import coerce_multi_time_kernel  # NOQA
