"""
Moved to:
    ~/code/kwcoco_dataloader/kwcoco_dataloader/tasks/fusion/datamodules/temporal_sampling/time_kernel.py
"""
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import TimeKernel  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.temporal_sampling.affinity import _random_discrete_relative_times  # NOQA
