# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: deeplab2/dataset.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='deeplab2/dataset.proto',
    package='deeplab2',
    syntax='proto2',
    serialized_options=None,
    serialized_pb=b'\n\x16\x64\x65\x65plab2/dataset.proto\x12\x08\x64\x65\x65plab2\"\xc2\x05\n\x0e\x44\x61tasetOptions\x12\x0f\n\x07\x64\x61taset\x18\x01 \x01(\t\x12\x14\n\x0c\x66ile_pattern\x18\x02 \x03(\t\x12\x16\n\nbatch_size\x18\x03 \x01(\x05:\x02\x33\x32\x12\x11\n\tcrop_size\x18\x04 \x03(\x05\x12\x18\n\x10min_resize_value\x18\x05 \x03(\x05\x12\x18\n\x10max_resize_value\x18\x06 \x03(\x05\x12\x15\n\rresize_factor\x18\x07 \x01(\x05\x12\x43\n\raugmentations\x18\x08 \x01(\x0b\x32,.deeplab2.DatasetOptions.AugmentationOptions\x12\x10\n\x05sigma\x18\t \x01(\x02:\x01\x38\x12.\n\x1fincrease_small_instance_weights\x18\n \x01(\x08:\x05\x66\x61lse\x12&\n\x18small_instance_threshold\x18\x0b \x01(\x05:\x04\x34\x30\x39\x36\x12 \n\x15small_instance_weight\x18\x0c \x01(\x02:\x01\x33\x12\x1d\n\x0euse_two_frames\x18\r \x01(\x08:\x05\x66\x61lse\x12&\n\x18\x64\x65\x63ode_groundtruth_label\x18\x0e \x01(\x08:\x04true\x12(\n\x19thing_id_mask_annotations\x18\x0f \x01(\x08:\x05\x66\x61lse\x12\x19\n\x0cmax_thing_id\x18\x10 \x01(\x05:\x03\x31\x32\x38\x12\x1d\n\x0euse_next_frame\x18\x11 \x01(\x08:\x05\x66\x61lse\x1a\x96\x01\n\x13\x41ugmentationOptions\x12\x1b\n\x10min_scale_factor\x18\x01 \x01(\x02:\x01\x31\x12\x1b\n\x10max_scale_factor\x18\x02 \x01(\x02:\x01\x31\x12$\n\x16scale_factor_step_size\x18\x03 \x01(\x02:\x04\x30.25\x12\x1f\n\x17\x61utoaugment_policy_name\x18\x04 \x01(\t'
)

_DATASETOPTIONS_AUGMENTATIONOPTIONS = _descriptor.Descriptor(
    name='AugmentationOptions',
    full_name='deeplab2.DatasetOptions.AugmentationOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='min_scale_factor', full_name='deeplab2.DatasetOptions.AugmentationOptions.min_scale_factor', index=0,
            number=1, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_scale_factor', full_name='deeplab2.DatasetOptions.AugmentationOptions.max_scale_factor', index=1,
            number=2, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(1),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='scale_factor_step_size',
            full_name='deeplab2.DatasetOptions.AugmentationOptions.scale_factor_step_size', index=2,
            number=3, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(0.25),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='autoaugment_policy_name',
            full_name='deeplab2.DatasetOptions.AugmentationOptions.autoaugment_policy_name', index=3,
            number=4, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=593,
    serialized_end=743,
)

_DATASETOPTIONS = _descriptor.Descriptor(
    name='DatasetOptions',
    full_name='deeplab2.DatasetOptions',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='dataset', full_name='deeplab2.DatasetOptions.dataset', index=0,
            number=1, type=9, cpp_type=9, label=1,
            has_default_value=False, default_value=b"".decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='file_pattern', full_name='deeplab2.DatasetOptions.file_pattern', index=1,
            number=2, type=9, cpp_type=9, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='batch_size', full_name='deeplab2.DatasetOptions.batch_size', index=2,
            number=3, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=32,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='crop_size', full_name='deeplab2.DatasetOptions.crop_size', index=3,
            number=4, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='min_resize_value', full_name='deeplab2.DatasetOptions.min_resize_value', index=4,
            number=5, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_resize_value', full_name='deeplab2.DatasetOptions.max_resize_value', index=5,
            number=6, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='resize_factor', full_name='deeplab2.DatasetOptions.resize_factor', index=6,
            number=7, type=5, cpp_type=1, label=1,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='augmentations', full_name='deeplab2.DatasetOptions.augmentations', index=7,
            number=8, type=11, cpp_type=10, label=1,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='sigma', full_name='deeplab2.DatasetOptions.sigma', index=8,
            number=9, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(8),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='increase_small_instance_weights', full_name='deeplab2.DatasetOptions.increase_small_instance_weights',
            index=9,
            number=10, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='small_instance_threshold', full_name='deeplab2.DatasetOptions.small_instance_threshold', index=10,
            number=11, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=4096,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='small_instance_weight', full_name='deeplab2.DatasetOptions.small_instance_weight', index=11,
            number=12, type=2, cpp_type=6, label=1,
            has_default_value=True, default_value=float(3),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_two_frames', full_name='deeplab2.DatasetOptions.use_two_frames', index=12,
            number=13, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='decode_groundtruth_label', full_name='deeplab2.DatasetOptions.decode_groundtruth_label', index=13,
            number=14, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=True,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='thing_id_mask_annotations', full_name='deeplab2.DatasetOptions.thing_id_mask_annotations', index=14,
            number=15, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='max_thing_id', full_name='deeplab2.DatasetOptions.max_thing_id', index=15,
            number=16, type=5, cpp_type=1, label=1,
            has_default_value=True, default_value=128,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
        _descriptor.FieldDescriptor(
            name='use_next_frame', full_name='deeplab2.DatasetOptions.use_next_frame', index=16,
            number=17, type=8, cpp_type=7, label=1,
            has_default_value=True, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            serialized_options=None, file=DESCRIPTOR),
    ],
    extensions=[
    ],
    nested_types=[_DATASETOPTIONS_AUGMENTATIONOPTIONS, ],
    enum_types=[
    ],
    serialized_options=None,
    is_extendable=False,
    syntax='proto2',
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=37,
    serialized_end=743,
)

_DATASETOPTIONS_AUGMENTATIONOPTIONS.containing_type = _DATASETOPTIONS
_DATASETOPTIONS.fields_by_name['augmentations'].message_type = _DATASETOPTIONS_AUGMENTATIONOPTIONS
DESCRIPTOR.message_types_by_name['DatasetOptions'] = _DATASETOPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DatasetOptions = _reflection.GeneratedProtocolMessageType('DatasetOptions', (_message.Message,), {

    'AugmentationOptions': _reflection.GeneratedProtocolMessageType('AugmentationOptions', (_message.Message,), {
        'DESCRIPTOR': _DATASETOPTIONS_AUGMENTATIONOPTIONS,
        '__module__': 'deeplab2.dataset_pb2'
        # @@protoc_insertion_point(class_scope:deeplab2.DatasetOptions.AugmentationOptions)
    })
    ,
    'DESCRIPTOR': _DATASETOPTIONS,
    '__module__': 'deeplab2.dataset_pb2'
    # @@protoc_insertion_point(class_scope:deeplab2.DatasetOptions)
})
_sym_db.RegisterMessage(DatasetOptions)
_sym_db.RegisterMessage(DatasetOptions.AugmentationOptions)

# @@protoc_insertion_point(module_scope)
