# flake8: noqa
from .inria_aerial import InriaAerial
from .luxcarta_buildings import LuxcartaBuildings
from .mapping_challenge import MappingChallenge
from .open_cities_competition import RasterizedOpenCities, OpenCitiesTestDataset
from . import utils
from .xview2_dataset import xView2Dataset
